use std::time::Instant;

const BLOCK_SIZE: usize = 3;
const SIZE: usize = BLOCK_SIZE * BLOCK_SIZE;
const NUM_FIELDS: usize = SIZE * SIZE;

struct SField {
    field: Vec<u8>,
    fixed: Vec<bool>,
    pos: usize,
    possible_values: Vec<Vec<u8>>,
}

impl SField {
    pub fn new() -> SField {
        #[rustfmt::skip]
        let field = vec![
            0,0,0,0,0,0,0,0,0,
            0,0,0,0,8,9,0,0,0,
            0,0,0,0,2,0,4,0,0,
            0,0,4,0,6,0,0,0,8,
            0,0,0,5,0,0,0,0,0,
            0,6,5,0,0,2,0,7,4,
            0,3,0,0,0,5,0,4,0,
            0,0,1,8,0,0,0,0,0,
            0,0,8,2,0,0,0,6,0,
        ];

        let mut fixed = Vec::with_capacity(SIZE * SIZE);
        field.iter().for_each(|e| {
            if *e == 0 {
                fixed.push(false);
            } else {
                fixed.push(true);
            }
        });

        SField {
            field,
            fixed,
            pos: 0,
            possible_values: vec![vec![]; SIZE * SIZE],
        }
    }

    fn build_possible_values_db(&mut self) {
        for idx in 0..NUM_FIELDS {
            if self.fixed[idx] {
                continue;
            }
            self.pos = idx;
            // try all values between 1 and =self.size and remember the good ones
            let mut good_ones = Vec::with_capacity(SIZE);
            for nr in 1..=(SIZE as u8) {
                if self.ok(nr) {
                    good_ones.push(nr);
                }
            }
            self.possible_values[idx] = good_ones;
        }
        self.pos = 0;
    }

    pub fn solve_backtracking(&mut self) -> bool {
        let mut num_solutions = 0;
        loop {
            if self.is_end() {
                num_solutions += 1;

                if num_solutions % 10_000 == 0 {
                    self.print_clear();
                    self.print();
                    println!("Number of solutions: {}", num_solutions);
                }

                if self.is_fixed() {
                    continue;
                }
                self.clear_current_field();
                self.prev();
                self.goto_prev_free_field();
            }

            if self.is_fixed() {
                self.next();
                continue;
            }

            if self.goto_next_free_field() {
                if self.put_valid_nr() {
                    self.next();
                    continue;
                } else {
                    self.clear_current_field();
                    if !self.prev() {
                        println!("Number of solutions: {}", num_solutions);
                        break;
                    }
                    if !self.goto_prev_free_field() {
                        println!("Number of solutions: {}", num_solutions);
                        break;
                    }
                }
            }
        }

        num_solutions > 0
    }

    #[inline(always)]
    fn print_gray(&self) {
        print!("\x1b\x5b\x31\x3b\x33\x30\x6d");
    }

    #[inline(always)]
    fn print_green(&self) {
        print!("\x1b\x5b\x31\x3b\x33\x32\x6d");
    }

    #[inline(always)]
    fn print_neutral(&self) {
        print!("\x1b\x5b\x31\x3b\x30\x6d");
    }

    #[inline(always)]
    fn print_clear(&self) {
        print!("\x1b\x5b\x48\x1b\x5b\x32\x4a");
    }

    #[inline(always)]
    fn print(&self) {
        for i in 0..NUM_FIELDS {
            if i != 0 && i % SIZE == 0 {
                println!();
            }

            if i == self.pos {
                self.print_green();
            } else if self.get_field_at_pos(i) == 0 {
                self.print_gray();
            }

            print!("{:2} ", self.get_field_at_pos(i));

            if i == self.pos || self.get_field_at_pos(i) == 0 {
                self.print_neutral();
            }
        }
        println!();
    }

    #[inline(always)]
    fn clear_current_field(&mut self) {
        self.set(0);
    }

    #[inline(always)]
    fn goto_prev_free_field(&mut self) -> bool {
        while {
            if !self.is_fixed() {
                return true;
            }
            self.prev()
        } {}

        false
    }

    #[inline(always)]
    fn goto_next_free_field(&mut self) -> bool {
        while {
            if !self.is_fixed() {
                return true;
            }
            self.next()
        } {}

        false
    }

    #[inline(always)]
    fn _put_valid_nr(&mut self) -> bool {
        let current_nr = self.get_field_at_pos(self.pos);

        for nr in current_nr..(SIZE + 1) as u8 {
            if self.ok(nr) {
                self.set(nr);
                return true;
            }
        }

        false
    }

    #[inline(always)]
    fn put_valid_nr(&mut self) -> bool {
        let current_nr = self.get_field_at_pos(self.pos);

        // safety:
        //     self.possible_vals is initialized with self.size * self.size
        //     so self.pos can safely be used to index here
        let possible_vals = unsafe { self.possible_values.get_unchecked(self.pos) };

        for nr in possible_vals.iter() {
            if *nr <= current_nr {
                continue;
            }

            if self.ok(*nr) {
                self.set(*nr);
                return true;
            }
        }
        false
    }

    #[inline(always)]
    fn is_end(&self) -> bool {
        !self.field.contains(&0)
    }

    #[inline(always)]
    fn ok(&self, nr: u8) -> bool {
        self.block_ok(nr) && self.row_ok(nr) && self.col_ok(nr)
    }

    #[inline(always)]
    fn _pos_to_xy(&self) -> (usize, usize) {
        let y = self.pos / SIZE;
        let x = self.pos % SIZE;
        (x + 1, y + 1)
    }

    #[inline(always)]
    fn is_fixed(&self) -> bool {
        // safety: self.pos can be used to index the field unchecked
        //         since the only methods modifying self.pos are
        //         `next()` and `prev()` and they do bounds checking
        unsafe { *self.fixed.get_unchecked(self.pos) }
    }

    #[inline(always)]
    fn get_field_at_pos(&self, pos: usize) -> u8 {
        // safety:
        //     TODO
        //     would need to mathematically explain that the calculations in
        //     get_row(), get_col() and get_block() can never exceed 0..self.num_fields
        unsafe { *self.field.get_unchecked(pos) }
    }

    #[inline(always)]
    fn set(&mut self, nr: u8) {
        self.field[self.pos] = nr;
    }

    #[inline(always)]
    fn get_row(&self, row: &mut [u8; SIZE]) {
        for (idx, row_elem) in row.iter_mut().enumerate() {
            *row_elem = self.get_field_at_pos((self.pos / SIZE) * SIZE + idx);
        }
    }

    #[inline(always)]
    fn get_col(&self, col: &mut [u8; SIZE]) {
        for (idx, col_elem) in col.iter_mut().enumerate() {
            *col_elem = self.get_field_at_pos(idx * SIZE + self.pos % SIZE);
        }
    }

    #[inline(always)]
    fn get_block(&self, block: &mut [u8; SIZE]) {
        let block_start_row = self.pos / SIZE / BLOCK_SIZE * BLOCK_SIZE;
        let block_start_col = self.pos % SIZE / BLOCK_SIZE * BLOCK_SIZE;

        for r in 0..BLOCK_SIZE {
            for c in 0..BLOCK_SIZE {
                block[r * BLOCK_SIZE + c] =
                    self.get_field_at_pos(SIZE * (block_start_row + r) + block_start_col + c);
            }
        }
    }

    #[inline(always)]
    fn block_ok(&self, nr: u8) -> bool {
        let mut block = [0; SIZE];
        self.get_block(&mut block);
        !block.contains(&nr)
    }

    #[inline(always)]
    fn row_ok(&self, nr: u8) -> bool {
        let mut row = [0; SIZE];
        self.get_row(&mut row);
        !row.contains(&nr)
    }

    #[inline(always)]
    fn col_ok(&self, nr: u8) -> bool {
        let mut col = [0; SIZE];
        self.get_col(&mut col);
        !col.contains(&nr)
    }

    #[inline(always)]
    fn next(&mut self) -> bool {
        if self.pos == NUM_FIELDS - 1 {
            return false;
        }

        self.pos += 1;
        true
    }

    #[inline(always)]
    fn prev(&mut self) -> bool {
        if self.pos == 0 {
            return false;
        }

        self.pos -= 1;
        true
    }

    #[inline(always)]
    fn solve(&mut self) {
        let now = Instant::now();
        if !self.solve_backtracking() {
            println!("there is no solution.");
        }
        eprintln!("took {:.3}s", now.elapsed().as_secs_f64());
    }
}

fn run() -> Result<(), String> {
    let mut field = SField::new();

    field.build_possible_values_db();

    field.solve();
    Ok(())
}

fn main() {
    if let Err(e) = run() {
        println!("{}", e)
    }
}
